D2907: wireproto: add streams to frame-based protocol

indygreg (Gregory Szorc) phabricator at mercurial-scm.org
Tue Mar 20 21:41:34 UTC 2018


indygreg created this revision.
Herald added a subscriber: mercurial-devel.
Herald added a reviewer: hg-reviewers.

REVISION SUMMARY
  Previously, the frame-based protocol was just a series of frames,
  with each frame associated with a request ID.
  
  In order to scale the protocol, we'll want to enable the use of
  compression. While it is possible to enable compression at the
  socket/pipe level, this has its disadvantages. The big one is it
  undermines the point of frames being standalone, atomic units that
  can be read and written: if you add compression above the framing
  protocol, you are back to having a stream-based protocol as opposed
  to something frame-based.
  
  So in order to preserve frames, compression needs to occur at
  the frame payload level.
  
  Compressing each frame's payload individually will limit compression
  ratios because the window size of the compressor will be limited
  by the max frame size, which is 32-64kb as currently defined. It
  will also add CPU overhead, as it is more efficient for compressors
  to operate on fewer, larger blocks of data than more, smaller blocks.
  
  So compressing each frame independently is out.
  
  This means we need to compress each frame's payload as if it is part
  of a larger stream.
  
  The simplest approach is to have 1 stream per connection. This
  could certainly work. However, it has disadvantages (documented below).
  
  We could also have 1 stream per RPC/command invocation. (This is the
  model HTTP/2 goes with.) This also has disadvantages.
  
  The main disadvantage to one global stream is that it has the very
  real potential to create CPU bottlenecks doing compression. Networks
  are only getting faster and the performance of single CPU cores has
  been relatively flat. Newer compression formats like zstandard offer
  better CPU cycle efficiency than predecessors like zlib. But it still
  all too common to saturate your CPU with compression overhead long
  before you saturate the network pipe.
  
  The main disadvantage with streams per request is that you can't
  reap the benefits of the compression context for multiple requests.
  For example, if you send 1000 RPC requests (or HTTP/2 requests for
  that matter), the response to each would have its own compression
  context. The overall size of the raw responses would be larger because
  compression contexts wouldn't be able to reference data from another
  request or response.
  
  The approach for streams as implemented in this commit is to support
  N streams per connection and for streams to potentially span requests
  and responses. As explained by the added internals docs, this
  facilitates servers and clients delegating independent streams and
  compression to independent threads / CPU cores. This helps alleviate
  the CPU bottleneck of compression. This design also allows compression
  contexts to be reused across requests/responses. This can result in
  improved compression ratios and less overhead for compressors and
  decompressors having to build new contexts.
  
  Another feature that was defined was the ability for individual frames
  within a stream to declare whether that individual frame's payload
  uses the content encoding (read: compression) defined by the stream.
  The idea here is that some servers may serve data from a combination
  of caches and dynamic resolution. Data coming from caches may be
  pre-compressed. We want to facilitate servers being able to essentially
  stream bytes from caches to the wire with minimal overhead. Being
  able to mix and match with frames are compressed within a stream
  enables these types of advanced server functionality.
  
  This commit defines the new streams mechanism. Basic code for
  supporting streams in frames has been added. But that code is
  seriously lacking and doesn't fully conform to the defined protocol.
  For example, we don't close any streams. And support for content
  encoding within streams is not yet implemented. The change was
  rather invasive and I didn't think it would be reasonable to implement
  the entire feature in a single commit.
  
  For the record, I would have loved to reuse an existing multiplexing
  protocol to build the new wire protocol on top of. However, I couldn't
  find a protocol that offers the performance and scaling characteristics
  that I desired. Namely, it should support multiple compression
  contexts to facilitate scaling out to multiple CPU cores and
  compression contexts should be able to live longer than single RPC
  requests. HTTP/2 *almost* fits the bill. But the semantics of HTTP
  message exchange state that streams can only live for a single
  request-response. We /could/ tunnel on top of HTTP/2 streams and
  frames with HEADER and DATA frames. But there's no guarantee that
  HTTP/2 libraries and proxies would allow us to use HTTP/2 streams
  and frames without the HTTP message exchange semantics defined in
  RFC 7540 Section 8. Other RPC protocols like gRPC tunnel are built
  on top of HTTP/2 and thus preserve its semantics of stream per
  RPC invocation. Even QUIC does this. We could attempt to invent a
  higher-level stream that spans HTTP/2 streams. But this would be
  violating HTTP/2 because there is no guarantee that HTTP/2 streams
  are routed to the same server. The best we can do - which is what
  this protocol does - is shoehorn all request and response data into
  a single HTTP message and create streams within. At that point, we've
  defined a Content-Type in HTTP parlance. It just so happens our
  media type can also work as a standalone, stream-based protocol,
  without leaning on HTTP or similar protocol.

REPOSITORY
  rHG Mercurial

REVISION DETAIL
  https://phab.mercurial-scm.org/D2907

AFFECTED FILES
  mercurial/help/internals/wireprotocol.txt
  mercurial/wireprotoframing.py
  mercurial/wireprotoserver.py
  tests/test-http-api-httpv2.t
  tests/test-wireproto-serverreactor.py

CHANGE DETAILS

diff --git a/tests/test-wireproto-serverreactor.py b/tests/test-wireproto-serverreactor.py
--- a/tests/test-wireproto-serverreactor.py
+++ b/tests/test-wireproto-serverreactor.py
@@ -23,6 +23,8 @@
         assert len(payload) == header.length
 
         yield reactor.onframerecv(framing.frame(header.requestid,
+                                                header.streamid,
+                                                header.streamflags,
                                                 header.typeid,
                                                 header.flags,
                                                 payload))
@@ -37,44 +39,44 @@
     def testdataexactframesize(self):
         data = util.bytesio(b'x' * framing.DEFAULT_MAX_FRAME_SIZE)
 
-        stream = framing.stream()
+        stream = framing.stream(1)
         frames = list(framing.createcommandframes(stream, 1, b'command',
                                                   {}, data))
         self.assertEqual(frames, [
-            ffs(b'1 command-name have-data command'),
-            ffs(b'1 command-data continuation %s' % data.getvalue()),
-            ffs(b'1 command-data eos ')
+            ffs(b'1 1 stream-begin command-name have-data command'),
+            ffs(b'1 1 0 command-data continuation %s' % data.getvalue()),
+            ffs(b'1 1 0 command-data eos ')
         ])
 
     def testdatamultipleframes(self):
         data = util.bytesio(b'x' * (framing.DEFAULT_MAX_FRAME_SIZE + 1))
 
-        stream = framing.stream()
+        stream = framing.stream(1)
         frames = list(framing.createcommandframes(stream, 1, b'command', {},
                                                   data))
         self.assertEqual(frames, [
-            ffs(b'1 command-name have-data command'),
-            ffs(b'1 command-data continuation %s' % (
+            ffs(b'1 1 stream-begin command-name have-data command'),
+            ffs(b'1 1 0 command-data continuation %s' % (
                 b'x' * framing.DEFAULT_MAX_FRAME_SIZE)),
-            ffs(b'1 command-data eos x'),
+            ffs(b'1 1 0 command-data eos x'),
         ])
 
     def testargsanddata(self):
         data = util.bytesio(b'x' * 100)
 
-        stream = framing.stream()
+        stream = framing.stream(1)
         frames = list(framing.createcommandframes(stream, 1, b'command', {
             b'key1': b'key1value',
             b'key2': b'key2value',
             b'key3': b'key3value',
         }, data))
 
         self.assertEqual(frames, [
-            ffs(b'1 command-name have-args|have-data command'),
-            ffs(br'1 command-argument 0 \x04\x00\x09\x00key1key1value'),
-            ffs(br'1 command-argument 0 \x04\x00\x09\x00key2key2value'),
-            ffs(br'1 command-argument eoa \x04\x00\x09\x00key3key3value'),
-            ffs(b'1 command-data eos %s' % data.getvalue()),
+            ffs(b'1 1 stream-begin command-name have-args|have-data command'),
+            ffs(br'1 1 0 command-argument 0 \x04\x00\x09\x00key1key1value'),
+            ffs(br'1 1 0 command-argument 0 \x04\x00\x09\x00key2key2value'),
+            ffs(br'1 1 0 command-argument eoa \x04\x00\x09\x00key3key3value'),
+            ffs(b'1 1 0 command-data eos %s' % data.getvalue()),
         ])
 
     def testtextoutputexcessiveargs(self):
@@ -128,64 +130,67 @@
                 (b'bleh', [], [b'x' * 65536])]))
 
     def testtextoutput1simpleatom(self):
-        stream = framing.stream()
+        stream = framing.stream(1)
         val = list(framing.createtextoutputframe(stream, 1, [
             (b'foo', [], [])]))
 
         self.assertEqual(val, [
-            ffs(br'1 text-output 0 \x03\x00\x00\x00foo'),
+            ffs(br'1 1 stream-begin text-output 0 \x03\x00\x00\x00foo'),
         ])
 
     def testtextoutput2simpleatoms(self):
-        stream = framing.stream()
+        stream = framing.stream(1)
         val = list(framing.createtextoutputframe(stream, 1, [
             (b'foo', [], []),
             (b'bar', [], []),
         ]))
 
         self.assertEqual(val, [
-            ffs(br'1 text-output 0 \x03\x00\x00\x00foo\x03\x00\x00\x00bar'),
+            ffs(br'1 1 stream-begin text-output 0 '
+                br'\x03\x00\x00\x00foo\x03\x00\x00\x00bar'),
         ])
 
     def testtextoutput1arg(self):
-        stream = framing.stream()
+        stream = framing.stream(1)
         val = list(framing.createtextoutputframe(stream, 1, [
             (b'foo %s', [b'val1'], []),
         ]))
 
         self.assertEqual(val, [
-            ffs(br'1 text-output 0 \x06\x00\x00\x01\x04\x00foo %sval1'),
+            ffs(br'1 1 stream-begin text-output 0 '
+                br'\x06\x00\x00\x01\x04\x00foo %sval1'),
         ])
 
     def testtextoutput2arg(self):
-        stream = framing.stream()
+        stream = framing.stream(1)
         val = list(framing.createtextoutputframe(stream, 1, [
             (b'foo %s %s', [b'val', b'value'], []),
         ]))
 
         self.assertEqual(val, [
-            ffs(br'1 text-output 0 \x09\x00\x00\x02\x03\x00\x05\x00'
-                br'foo %s %svalvalue'),
+            ffs(br'1 1 stream-begin text-output 0 '
+                br'\x09\x00\x00\x02\x03\x00\x05\x00foo %s %svalvalue'),
         ])
 
     def testtextoutput1label(self):
-        stream = framing.stream()
+        stream = framing.stream(1)
         val = list(framing.createtextoutputframe(stream, 1, [
             (b'foo', [], [b'label']),
         ]))
 
         self.assertEqual(val, [
-            ffs(br'1 text-output 0 \x03\x00\x01\x00\x05foolabel'),
+            ffs(br'1 1 stream-begin text-output 0 \x03\x00\x01\x00\x05foolabel'),
         ])
 
     def testargandlabel(self):
-        stream = framing.stream()
+        stream = framing.stream(1)
         val = list(framing.createtextoutputframe(stream, 1, [
             (b'foo %s', [b'arg'], [b'label']),
         ]))
 
         self.assertEqual(val, [
-            ffs(br'1 text-output 0 \x06\x00\x01\x01\x05\x03\x00foo %slabelarg'),
+            ffs(br'1 1 stream-begin text-output 0 '
+                br'\x06\x00\x01\x01\x05\x03\x00foo %slabelarg'),
         ])
 
 class ServerReactorTests(unittest.TestCase):
@@ -208,7 +213,7 @@
     def test1framecommand(self):
         """Receiving a command in a single frame yields request to run it."""
         reactor = makereactor()
-        stream = framing.stream()
+        stream = framing.stream(1)
         results = list(sendcommandframes(reactor, stream, 1, b'mycommand', {}))
         self.assertEqual(len(results), 1)
         self.assertaction(results[0], 'runcommand')
@@ -224,7 +229,7 @@
 
     def test1argument(self):
         reactor = makereactor()
-        stream = framing.stream()
+        stream = framing.stream(1)
         results = list(sendcommandframes(reactor, stream, 41, b'mycommand',
                                          {b'foo': b'bar'}))
         self.assertEqual(len(results), 2)
@@ -239,7 +244,7 @@
 
     def testmultiarguments(self):
         reactor = makereactor()
-        stream = framing.stream()
+        stream = framing.stream(1)
         results = list(sendcommandframes(reactor, stream, 1, b'mycommand',
                                          {b'foo': b'bar', b'biz': b'baz'}))
         self.assertEqual(len(results), 3)
@@ -255,7 +260,7 @@
 
     def testsimplecommanddata(self):
         reactor = makereactor()
-        stream = framing.stream()
+        stream = framing.stream(1)
         results = list(sendcommandframes(reactor, stream, 1, b'mycommand', {},
                                          util.bytesio(b'data!')))
         self.assertEqual(len(results), 2)
@@ -270,10 +275,10 @@
 
     def testmultipledataframes(self):
         frames = [
-            ffs(b'1 command-name have-data mycommand'),
-            ffs(b'1 command-data continuation data1'),
-            ffs(b'1 command-data continuation data2'),
-            ffs(b'1 command-data eos data3'),
+            ffs(b'1 1 stream-begin command-name have-data mycommand'),
+            ffs(b'1 1 0 command-data continuation data1'),
+            ffs(b'1 1 0 command-data continuation data2'),
+            ffs(b'1 1 0 command-data eos data3'),
         ]
 
         reactor = makereactor()
@@ -291,11 +296,11 @@
 
     def testargumentanddata(self):
         frames = [
-            ffs(b'1 command-name have-args|have-data command'),
-            ffs(br'1 command-argument 0 \x03\x00\x03\x00keyval'),
-            ffs(br'1 command-argument eoa \x03\x00\x03\x00foobar'),
-            ffs(b'1 command-data continuation value1'),
-            ffs(b'1 command-data eos value2'),
+            ffs(b'1 1 stream-begin command-name have-args|have-data command'),
+            ffs(br'1 1 0 command-argument 0 \x03\x00\x03\x00keyval'),
+            ffs(br'1 1 0 command-argument eoa \x03\x00\x03\x00foobar'),
+            ffs(b'1 1 0 command-data continuation value1'),
+            ffs(b'1 1 0 command-data eos value2'),
         ]
 
         reactor = makereactor()
@@ -315,17 +320,17 @@
     def testunexpectedcommandargument(self):
         """Command argument frame when not running a command is an error."""
         result = self._sendsingleframe(makereactor(),
-                                       b'1 command-argument 0 ignored')
+                                       b'1 1 stream-begin command-argument 0 ignored')
         self.assertaction(result, 'error')
         self.assertEqual(result[1], {
             'message': b'expected command frame; got 2',
         })
 
     def testunexpectedcommandargumentreceiving(self):
         """Same as above but the command is receiving."""
         results = list(sendframes(makereactor(), [
-            ffs(b'1 command-name have-data command'),
-            ffs(b'1 command-argument eoa ignored'),
+            ffs(b'1 1 stream-begin command-name have-data command'),
+            ffs(b'1 1 0 command-argument eoa ignored'),
         ]))
 
         self.assertaction(results[1], 'error')
@@ -337,17 +342,17 @@
     def testunexpectedcommanddata(self):
         """Command argument frame when not running a command is an error."""
         result = self._sendsingleframe(makereactor(),
-                                       b'1 command-data 0 ignored')
+                                       b'1 1 stream-begin command-data 0 ignored')
         self.assertaction(result, 'error')
         self.assertEqual(result[1], {
             'message': b'expected command frame; got 3',
         })
 
     def testunexpectedcommanddatareceiving(self):
         """Same as above except the command is receiving."""
         results = list(sendframes(makereactor(), [
-            ffs(b'1 command-name have-args command'),
-            ffs(b'1 command-data eos ignored'),
+            ffs(b'1 1 stream-begin command-name have-args command'),
+            ffs(b'1 1 0 command-data eos ignored'),
         ]))
 
         self.assertaction(results[1], 'error')
@@ -359,7 +364,7 @@
     def testmissingcommandframeflags(self):
         """Command name frame must have flags set."""
         result = self._sendsingleframe(makereactor(),
-                                       b'1 command-name 0 command')
+                                       b'1 1 stream-begin command-name 0 command')
         self.assertaction(result, 'error')
         self.assertEqual(result[1], {
             'message': b'missing frame flags on command frame',
@@ -384,8 +389,8 @@
     def testconflictingrequestid(self):
         """Request ID for new command matching in-flight command is illegal."""
         results = list(sendframes(makereactor(), [
-            ffs(b'1 command-name have-args command'),
-            ffs(b'1 command-name eos command'),
+            ffs(b'1 1 stream-begin command-name have-args command'),
+            ffs(b'1 1 0 command-name eos command'),
         ]))
 
         self.assertaction(results[0], 'wantframe')
@@ -396,12 +401,12 @@
 
     def testinterleavedcommands(self):
         results = list(sendframes(makereactor(), [
-            ffs(b'1 command-name have-args command1'),
-            ffs(b'3 command-name have-args command3'),
-            ffs(br'1 command-argument 0 \x03\x00\x03\x00foobar'),
-            ffs(br'3 command-argument 0 \x03\x00\x03\x00bizbaz'),
-            ffs(br'3 command-argument eoa \x03\x00\x03\x00keyval'),
-            ffs(br'1 command-argument eoa \x04\x00\x03\x00key1val'),
+            ffs(b'1 1 stream-begin command-name have-args command1'),
+            ffs(b'3 1 0 command-name have-args command3'),
+            ffs(br'1 1 0 command-argument 0 \x03\x00\x03\x00foobar'),
+            ffs(br'3 1 0 command-argument 0 \x03\x00\x03\x00bizbaz'),
+            ffs(br'3 1 0 command-argument eoa \x03\x00\x03\x00keyval'),
+            ffs(br'1 1 0 command-argument eoa \x04\x00\x03\x00key1val'),
         ]))
 
         self.assertEqual([t[0] for t in results], [
@@ -431,17 +436,17 @@
         # command request waiting on argument data. But it doesn't handle that
         # scenario yet. So this test does nothing of value.
         frames = [
-            ffs(b'1 command-name have-args command'),
+            ffs(b'1 1 stream-begin command-name have-args command'),
         ]
 
         results = list(sendframes(makereactor(), frames))
         self.assertaction(results[0], 'wantframe')
 
     def testincompleteargumentname(self):
         """Argument frame with incomplete name."""
         frames = [
-            ffs(b'1 command-name have-args command1'),
-            ffs(br'1 command-argument eoa \x04\x00\xde\xadfoo'),
+            ffs(b'1 1 stream-begin command-name have-args command1'),
+            ffs(br'1 1 0 command-argument eoa \x04\x00\xde\xadfoo'),
         ]
 
         results = list(sendframes(makereactor(), frames))
@@ -455,8 +460,8 @@
     def testincompleteargumentvalue(self):
         """Argument frame with incomplete value."""
         frames = [
-            ffs(b'1 command-name have-args command'),
-            ffs(br'1 command-argument eoa \x03\x00\xaa\xaafoopartialvalue'),
+            ffs(b'1 1 stream-begin command-name have-args command'),
+            ffs(br'1 1 0 command-argument eoa \x03\x00\xaa\xaafoopartialvalue'),
         ]
 
         results = list(sendframes(makereactor(), frames))
@@ -471,18 +476,18 @@
         # The reactor doesn't currently handle partially received commands.
         # So this test is failing to do anything with request 1.
         frames = [
-            ffs(b'1 command-name have-data command1'),
-            ffs(b'3 command-name eos command2'),
+            ffs(b'1 1 stream-begin command-name have-data command1'),
+            ffs(b'3 1 0 command-name eos command2'),
         ]
         results = list(sendframes(makereactor(), frames))
         self.assertEqual(len(results), 2)
         self.assertaction(results[0], 'wantframe')
         self.assertaction(results[1], 'runcommand')
 
     def testmissingcommanddataframeflags(self):
         frames = [
-            ffs(b'1 command-name have-data command1'),
-            ffs(b'1 command-data 0 data'),
+            ffs(b'1 1 stream-begin command-name have-data command1'),
+            ffs(b'1 1 0 command-data 0 data'),
         ]
         results = list(sendframes(makereactor(), frames))
         self.assertEqual(len(results), 2)
@@ -495,9 +500,9 @@
     def testframefornonreceivingrequest(self):
         """Receiving a frame for a command that is not receiving is illegal."""
         results = list(sendframes(makereactor(), [
-            ffs(b'1 command-name eos command1'),
-            ffs(b'3 command-name have-data command3'),
-            ffs(b'5 command-argument eoa ignored'),
+            ffs(b'1 1 stream-begin command-name eos command1'),
+            ffs(b'3 1 0 command-name have-data command3'),
+            ffs(b'5 1 0 command-argument eoa ignored'),
         ]))
         self.assertaction(results[2], 'error')
         self.assertEqual(results[2][1], {
@@ -507,105 +512,105 @@
     def testsimpleresponse(self):
         """Bytes response to command sends result frames."""
         reactor = makereactor()
-        instream = framing.stream()
+        instream = framing.stream(1)
         list(sendcommandframes(reactor, instream, 1, b'mycommand', {}))
 
-        outstream = framing.stream()
+        outstream = framing.stream(2)
         result = reactor.onbytesresponseready(outstream, 1, b'response')
         self.assertaction(result, 'sendframes')
         self.assertframesequal(result[1]['framegen'], [
-            b'1 bytes-response eos response',
+            b'1 2 stream-begin bytes-response eos response',
         ])
 
     def testmultiframeresponse(self):
         """Bytes response spanning multiple frames is handled."""
         first = b'x' * framing.DEFAULT_MAX_FRAME_SIZE
         second = b'y' * 100
 
         reactor = makereactor()
-        instream = framing.stream()
+        instream = framing.stream(1)
         list(sendcommandframes(reactor, instream, 1, b'mycommand', {}))
 
-        outstream = framing.stream()
+        outstream = framing.stream(2)
         result = reactor.onbytesresponseready(outstream, 1, first + second)
         self.assertaction(result, 'sendframes')
         self.assertframesequal(result[1]['framegen'], [
-            b'1 bytes-response continuation %s' % first,
-            b'1 bytes-response eos %s' % second,
+            b'1 2 stream-begin bytes-response continuation %s' % first,
+            b'1 2 0 bytes-response eos %s' % second,
         ])
 
     def testapplicationerror(self):
         reactor = makereactor()
-        instream = framing.stream()
+        instream = framing.stream(1)
         list(sendcommandframes(reactor, instream, 1, b'mycommand', {}))
 
-        outstream = framing.stream()
+        outstream = framing.stream(2)
         result = reactor.onapplicationerror(outstream, 1, b'some message')
         self.assertaction(result, 'sendframes')
         self.assertframesequal(result[1]['framegen'], [
-            b'1 error-response application some message',
+            b'1 2 stream-begin error-response application some message',
         ])
 
     def test1commanddeferresponse(self):
         """Responses when in deferred output mode are delayed until EOF."""
         reactor = makereactor(deferoutput=True)
-        instream = framing.stream()
+        instream = framing.stream(1)
         results = list(sendcommandframes(reactor, instream, 1, b'mycommand', {}))
         self.assertEqual(len(results), 1)
         self.assertaction(results[0], 'runcommand')
 
-        outstream = framing.stream()
+        outstream = framing.stream(2)
         result = reactor.onbytesresponseready(outstream, 1, b'response')
         self.assertaction(result, 'noop')
         result = reactor.oninputeof()
         self.assertaction(result, 'sendframes')
         self.assertframesequal(result[1]['framegen'], [
-            b'1 bytes-response eos response',
+            b'1 2 stream-begin bytes-response eos response',
         ])
 
     def testmultiplecommanddeferresponse(self):
         reactor = makereactor(deferoutput=True)
-        instream = framing.stream()
+        instream = framing.stream(1)
         list(sendcommandframes(reactor, instream, 1, b'command1', {}))
         list(sendcommandframes(reactor, instream, 3, b'command2', {}))
 
-        outstream = framing.stream()
+        outstream = framing.stream(2)
         result = reactor.onbytesresponseready(outstream, 1, b'response1')
         self.assertaction(result, 'noop')
         result = reactor.onbytesresponseready(outstream, 3, b'response2')
         self.assertaction(result, 'noop')
         result = reactor.oninputeof()
         self.assertaction(result, 'sendframes')
         self.assertframesequal(result[1]['framegen'], [
-            b'1 bytes-response eos response1',
-            b'3 bytes-response eos response2'
+            b'1 2 stream-begin bytes-response eos response1',
+            b'3 2 0 bytes-response eos response2'
         ])
 
     def testrequestidtracking(self):
         reactor = makereactor(deferoutput=True)
-        instream = framing.stream()
+        instream = framing.stream(1)
         list(sendcommandframes(reactor, instream, 1, b'command1', {}))
         list(sendcommandframes(reactor, instream, 3, b'command2', {}))
         list(sendcommandframes(reactor, instream, 5, b'command3', {}))
 
         # Register results for commands out of order.
-        outstream = framing.stream()
+        outstream = framing.stream(2)
         reactor.onbytesresponseready(outstream, 3, b'response3')
         reactor.onbytesresponseready(outstream, 1, b'response1')
         reactor.onbytesresponseready(outstream, 5, b'response5')
 
         result = reactor.oninputeof()
         self.assertaction(result, 'sendframes')
         self.assertframesequal(result[1]['framegen'], [
-            b'3 bytes-response eos response3',
-            b'1 bytes-response eos response1',
-            b'5 bytes-response eos response5',
+            b'3 2 stream-begin bytes-response eos response3',
+            b'1 2 0 bytes-response eos response1',
+            b'5 2 0 bytes-response eos response5',
         ])
 
     def testduplicaterequestonactivecommand(self):
         """Receiving a request ID that matches a request that isn't finished."""
         reactor = makereactor()
-        stream = framing.stream()
+        stream = framing.stream(1)
         list(sendcommandframes(reactor, stream, 1, b'command1', {}))
         results = list(sendcommandframes(reactor, stream, 1, b'command1', {}))
 
@@ -617,9 +622,9 @@
     def testduplicaterequestonactivecommandnosend(self):
         """Same as above but we've registered a response but haven't sent it."""
         reactor = makereactor()
-        instream = framing.stream()
+        instream = framing.stream(1)
         list(sendcommandframes(reactor, instream, 1, b'command1', {}))
-        outstream = framing.stream()
+        outstream = framing.stream(2)
         reactor.onbytesresponseready(outstream, 1, b'response')
 
         # We've registered the response but haven't sent it. From the
@@ -634,11 +639,11 @@
     def testduplicaterequestargumentframe(self):
         """Variant on above except we sent an argument frame instead of name."""
         reactor = makereactor()
-        stream = framing.stream()
+        stream = framing.stream(1)
         list(sendcommandframes(reactor, stream, 1, b'command', {}))
         results = list(sendframes(reactor, [
-            ffs(b'3 command-name have-args command'),
-            ffs(b'1 command-argument 0 ignored'),
+            ffs(b'3 1 stream-begin command-name have-args command'),
+            ffs(b'1 1 0 command-argument 0 ignored'),
         ]))
         self.assertaction(results[0], 'wantframe')
         self.assertaction(results[1], 'error')
@@ -649,9 +654,9 @@
     def testduplicaterequestaftersend(self):
         """We can use a duplicate request ID after we've sent the response."""
         reactor = makereactor()
-        instream = framing.stream()
+        instream = framing.stream(1)
         list(sendcommandframes(reactor, instream, 1, b'command1', {}))
-        outstream = framing.stream()
+        outstream = framing.stream(2)
         res = reactor.onbytesresponseready(outstream, 1, b'response')
         list(res[1]['framegen'])
 
diff --git a/tests/test-http-api-httpv2.t b/tests/test-http-api-httpv2.t
--- a/tests/test-http-api-httpv2.t
+++ b/tests/test-http-api-httpv2.t
@@ -179,7 +179,7 @@
   >     accept: $MEDIATYPE
   >     content-type: $MEDIATYPE
   >     user-agent: test
-  >     frame 1 command-name eos customreadonly
+  >     frame 1 1 stream-begin command-name eos customreadonly
   > EOF
   using raw connection to peer
   s>     POST /api/exp-http-v2-0001/ro/customreadonly HTTP/1.1\r\n
@@ -190,16 +190,16 @@
   s>     *\r\n (glob)
   s>     host: $LOCALIP:$HGPORT\r\n (glob)
   s>     \r\n
-  s>     \x0e\x00\x00\x01\x00\x11customreadonly
+  s>     \x0e\x00\x00\x01\x00\x01\x01\x11customreadonly
   s> makefile('rb', None)
   s>     HTTP/1.1 200 OK\r\n
   s>     Server: testing stub value\r\n
   s>     Date: $HTTP_DATE$\r\n
   s>     Content-Type: application/mercurial-exp-framing-0002\r\n
   s>     Transfer-Encoding: chunked\r\n
   s>     \r\n
-  s>     23\r\n
-  s>     \x1d\x00\x00\x01\x00Bcustomreadonly bytes response
+  s>     25\r\n
+  s>     \x1d\x00\x00\x01\x00\x02\x01Bcustomreadonly bytes response
   s>     \r\n
   s>     0\r\n
   s>     \r\n
@@ -290,27 +290,27 @@
   >     user-agent: test
   >     accept: $MEDIATYPE
   >     content-type: $MEDIATYPE
-  >     frame 1 command-name eos customreadonly
+  >     frame 1 1 stream-begin command-name eos customreadonly
   > EOF
   using raw connection to peer
   s>     POST /api/exp-http-v2-0001/rw/customreadonly HTTP/1.1\r\n
   s>     Accept-Encoding: identity\r\n
   s>     accept: application/mercurial-exp-framing-0002\r\n
   s>     content-type: application/mercurial-exp-framing-0002\r\n
   s>     user-agent: test\r\n
-  s>     content-length: 20\r\n
+  s>     content-length: 22\r\n
   s>     host: $LOCALIP:$HGPORT\r\n (glob)
   s>     \r\n
-  s>     \x0e\x00\x00\x01\x00\x11customreadonly
+  s>     \x0e\x00\x00\x01\x00\x01\x01\x11customreadonly
   s> makefile('rb', None)
   s>     HTTP/1.1 200 OK\r\n
   s>     Server: testing stub value\r\n
   s>     Date: $HTTP_DATE$\r\n
   s>     Content-Type: application/mercurial-exp-framing-0002\r\n
   s>     Transfer-Encoding: chunked\r\n
   s>     \r\n
-  s>     23\r\n
-  s>     \x1d\x00\x00\x01\x00Bcustomreadonly bytes response
+  s>     25\r\n
+  s>     \x1d\x00\x00\x01\x00\x02\x01Bcustomreadonly bytes response
   s>     \r\n
   s>     0\r\n
   s>     \r\n
@@ -382,20 +382,20 @@
   >     accept: $MEDIATYPE
   >     content-type: $MEDIATYPE
   >     user-agent: test
-  >     frame 1 command-name have-args command1
-  >     frame 1 command-argument 0 \x03\x00\x04\x00fooval1
-  >     frame 1 command-argument eoa \x04\x00\x03\x00bar1val
+  >     frame 1 1 stream-begin command-name have-args command1
+  >     frame 1 1 0 command-argument 0 \x03\x00\x04\x00fooval1
+  >     frame 1 1 0 command-argument eoa \x04\x00\x03\x00bar1val
   > EOF
   using raw connection to peer
   s>     POST /api/exp-http-v2-0001/ro/debugreflect HTTP/1.1\r\n
   s>     Accept-Encoding: identity\r\n
   s>     accept: application/mercurial-exp-framing-0002\r\n
   s>     content-type: application/mercurial-exp-framing-0002\r\n
   s>     user-agent: test\r\n
-  s>     content-length: 48\r\n
+  s>     content-length: 54\r\n
   s>     host: $LOCALIP:$HGPORT\r\n (glob)
   s>     \r\n
-  s>     \x08\x00\x00\x01\x00\x12command1\x0b\x00\x00\x01\x00 \x03\x00\x04\x00fooval1\x0b\x00\x00\x01\x00"\x04\x00\x03\x00bar1val
+  s>     \x08\x00\x00\x01\x00\x01\x01\x12command1\x0b\x00\x00\x01\x00\x01\x00 \x03\x00\x04\x00fooval1\x0b\x00\x00\x01\x00\x01\x00"\x04\x00\x03\x00bar1val
   s> makefile('rb', None)
   s>     HTTP/1.1 200 OK\r\n
   s>     Server: testing stub value\r\n
@@ -419,19 +419,19 @@
   >     accept: $MEDIATYPE
   >     content-type: $MEDIATYPE
   >     user-agent: test
-  >     frame 1 command-name eos customreadonly
-  >     frame 3 command-name eos customreadonly
+  >     frame 1 1 stream-begin command-name eos customreadonly
+  >     frame 3 1 0 command-name eos customreadonly
   > EOF
   using raw connection to peer
   s>     POST /api/exp-http-v2-0001/ro/customreadonly HTTP/1.1\r\n
   s>     Accept-Encoding: identity\r\n
   s>     accept: application/mercurial-exp-framing-0002\r\n
   s>     content-type: application/mercurial-exp-framing-0002\r\n
   s>     user-agent: test\r\n
-  s>     content-length: 40\r\n
+  s>     content-length: 44\r\n
   s>     host: $LOCALIP:$HGPORT\r\n (glob)
   s>     \r\n
-  s>     \x0e\x00\x00\x01\x00\x11customreadonly\x0e\x00\x00\x03\x00\x11customreadonly
+  s>     \x0e\x00\x00\x01\x00\x01\x01\x11customreadonly\x0e\x00\x00\x03\x00\x01\x00\x11customreadonly
   s> makefile('rb', None)
   s>     HTTP/1.1 200 OK\r\n
   s>     Server: testing stub value\r\n
@@ -448,8 +448,8 @@
   >     accept: $MEDIATYPE
   >     content-type: $MEDIATYPE
   >     user-agent: test
-  >     frame 1 command-name eos customreadonly
-  >     frame 3 command-name eos customreadonly
+  >     frame 1 1 stream-begin command-name eos customreadonly
+  >     frame 3 1 0 command-name eos customreadonly
   > EOF
   using raw connection to peer
   s>     POST /api/exp-http-v2-0001/ro/multirequest HTTP/1.1\r\n
@@ -460,19 +460,19 @@
   s>     *\r\n (glob)
   s>     host: $LOCALIP:$HGPORT\r\n (glob)
   s>     \r\n
-  s>     \x0e\x00\x00\x01\x00\x11customreadonly\x0e\x00\x00\x03\x00\x11customreadonly
+  s>     \x0e\x00\x00\x01\x00\x01\x01\x11customreadonly\x0e\x00\x00\x03\x00\x01\x00\x11customreadonly
   s> makefile('rb', None)
   s>     HTTP/1.1 200 OK\r\n
   s>     Server: testing stub value\r\n
   s>     Date: $HTTP_DATE$\r\n
   s>     Content-Type: application/mercurial-exp-framing-0002\r\n
   s>     Transfer-Encoding: chunked\r\n
   s>     \r\n
   s>     *\r\n (glob)
-  s>     \x1d\x00\x00\x01\x00Bcustomreadonly bytes response
+  s>     \x1d\x00\x00\x01\x00\x02\x01Bcustomreadonly bytes response
   s>     \r\n
-  s>     23\r\n
-  s>     \x1d\x00\x00\x03\x00Bcustomreadonly bytes response
+  s>     25\r\n
+  s>     \x1d\x00\x00\x03\x00\x02\x01Bcustomreadonly bytes response
   s>     \r\n
   s>     0\r\n
   s>     \r\n
@@ -484,34 +484,34 @@
   >     accept: $MEDIATYPE
   >     content-type: $MEDIATYPE
   >     user-agent: test
-  >     frame 1 command-name have-args listkeys
-  >     frame 3 command-name have-args listkeys
-  >     frame 3 command-argument eoa \x09\x00\x09\x00namespacebookmarks
-  >     frame 1 command-argument eoa \x09\x00\x0a\x00namespacenamespaces
+  >     frame 1 1 stream-begin command-name have-args listkeys
+  >     frame 3 1 0 command-name have-args listkeys
+  >     frame 3 1 0 command-argument eoa \x09\x00\x09\x00namespacebookmarks
+  >     frame 1 1 0 command-argument eoa \x09\x00\x0a\x00namespacenamespaces
   > EOF
   using raw connection to peer
   s>     POST /api/exp-http-v2-0001/ro/multirequest HTTP/1.1\r\n
   s>     Accept-Encoding: identity\r\n
   s>     accept: application/mercurial-exp-framing-0002\r\n
   s>     content-type: application/mercurial-exp-framing-0002\r\n
   s>     user-agent: test\r\n
-  s>     content-length: 85\r\n
+  s>     content-length: 93\r\n
   s>     host: $LOCALIP:$HGPORT\r\n (glob)
   s>     \r\n
-  s>     \x08\x00\x00\x01\x00\x12listkeys\x08\x00\x00\x03\x00\x12listkeys\x16\x00\x00\x03\x00"	\x00	\x00namespacebookmarks\x17\x00\x00\x01\x00"	\x00\n
+  s>     \x08\x00\x00\x01\x00\x01\x01\x12listkeys\x08\x00\x00\x03\x00\x01\x00\x12listkeys\x16\x00\x00\x03\x00\x01\x00"	\x00	\x00namespacebookmarks\x17\x00\x00\x01\x00\x01\x00"	\x00\n
   s>     \x00namespacenamespaces
   s> makefile('rb', None)
   s>     HTTP/1.1 200 OK\r\n
   s>     Server: testing stub value\r\n
   s>     Date: $HTTP_DATE$\r\n
   s>     Content-Type: application/mercurial-exp-framing-0002\r\n
   s>     Transfer-Encoding: chunked\r\n
   s>     \r\n
-  s>     6\r\n
-  s>     \x00\x00\x00\x03\x00B
+  s>     8\r\n
+  s>     \x00\x00\x00\x03\x00\x02\x01B
   s>     \r\n
-  s>     24\r\n
-  s>     \x1e\x00\x00\x01\x00Bbookmarks	\n
+  s>     26\r\n
+  s>     \x1e\x00\x00\x01\x00\x02\x01Bbookmarks	\n
   s>     namespaces	\n
   s>     phases	
   s>     \r\n
@@ -540,18 +540,18 @@
   >     accept: $MEDIATYPE
   >     content-type: $MEDIATYPE
   >     user-agent: test
-  >     frame 1 command-name eos unbundle
+  >     frame 1 1 stream-begin command-name eos unbundle
   > EOF
   using raw connection to peer
   s>     POST /api/exp-http-v2-0001/ro/multirequest HTTP/1.1\r\n
   s>     Accept-Encoding: identity\r\n
   s>     accept: application/mercurial-exp-framing-0002\r\n
   s>     content-type: application/mercurial-exp-framing-0002\r\n
   s>     user-agent: test\r\n
-  s>     content-length: 14\r\n
+  s>     content-length: 16\r\n
   s>     host: $LOCALIP:$HGPORT\r\n (glob)
   s>     \r\n
-  s>     \x08\x00\x00\x01\x00\x11unbundle
+  s>     \x08\x00\x00\x01\x00\x01\x01\x11unbundle
   s> makefile('rb', None)
   s>     HTTP/1.1 403 Forbidden\r\n
   s>     Server: testing stub value\r\n
diff --git a/mercurial/wireprotoserver.py b/mercurial/wireprotoserver.py
--- a/mercurial/wireprotoserver.py
+++ b/mercurial/wireprotoserver.py
@@ -541,7 +541,7 @@
 
     res.status = b'200 OK'
     res.headers[b'Content-Type'] = FRAMINGTYPE
-    stream = wireprotoframing.stream()
+    stream = wireprotoframing.stream(2)
 
     if isinstance(rsp, wireprototypes.bytesresponse):
         action, meta = reactor.onbytesresponseready(stream,
diff --git a/mercurial/help/internals/wireprotocol.txt b/mercurial/help/internals/wireprotocol.txt
--- a/mercurial/help/internals/wireprotocol.txt
+++ b/mercurial/help/internals/wireprotocol.txt
@@ -489,36 +489,47 @@
 ordered sends and receives is required. That is, each peer has one pipe
 for sending data and another for receiving.
 
+All data is read and written in atomic units called *frames*. These
+are conceptually similar to TCP packets. Higher-level functionality
+is built on the exchange and processing of frames.
+
+All frames are associated with a *stream*. A *stream* provides a
+unidirectional grouping of frames. Streams facilitate two goals:
+content encoding and parallelism. There is a dedicated section on
+streams below.
+
 The protocol is request-response based: the client issues requests to
 the server, which issues replies to those requests. Server-initiated
 messaging is not currently supported, but this specification carves
 out room to implement it.
 
-All data is read and written in atomic units called *frames*. These
-are conceptually similar to TCP packets. Higher-level functionality
-is built on the exchange and processing of frames.
-
 All frames are associated with a numbered request. Frames can thus
 be logically grouped by their request ID.
 
-Frames begin with a 6 octet header followed by a variable length
+Frames begin with an 8 octet header followed by a variable length
 payload::
 
-    +-----------------------------------------------+
-    |                 Length (24)                   |
-    +---------------------------------+-------------+
-    |         Request ID (16)         |
-    +----------+-----------+----------+
-    | Type (4) | Flags (4) |
-    +==========+===========+========================================|
+    +------------------------------------------------+
+    |                 Length (24)                    |
+    +--------------------------------+---------------+
+    |         Request ID (16)        | Stream ID (8) |
+    +------------------+-------------+---------------+
+    | Stream Flags (8) |
+    +-----------+------+
+    | Type (4)  |
+    +-----------+
+    | Flags (4) |
+    +===========+===================================================|
     |                     Frame Payload (0...)                    ...
     +---------------------------------------------------------------+
 
 The length of the frame payload is expressed as an unsigned 24 bit
 little endian integer. Values larger than 65535 MUST NOT be used unless
 given permission by the server as part of the negotiated capabilities
 during the handshake. The frame header is not part of the advertised
-frame length.
+frame length. The payload length is the over-the-wire length. If there
+is content encoding applied to the payload as part of the frame's stream,
+the length is the output of that content encoding, not the input.
 
 The 16-bit ``Request ID`` field denotes the integer request identifier,
 stored as an unsigned little endian integer. Odd numbered requests are
@@ -529,7 +540,16 @@
 start ordering request identifiers at ``1`` and ``0``, increment by
 ``2``, and wrap around if all available numbers have been exhausted.
 
-The 4-bit ``Type`` field denotes the type of message being sent.
+The 8-bit ``Stream ID`` field denotes the stream that the frame is
+associated with. Frames belonging to a stream may have content
+encoding applied and the receiver may need to decode the raw frame
+payload to obtain the original data. Odd numbered IDs are
+client-initiated. Even numbered IDs are server-initiated.
+
+The 8-bit ``Stream Flags`` field defines stream processing semantics.
+See the section on streams below.
+
+The 4-bit ``Type`` field denotes the type of frame being sent.
 
 The 4-bit ``Flags`` field defines special, per-type attributes for
 the frame.
@@ -758,6 +778,126 @@
 MUST normalize all string data to valid UTF-8 and receivers SHOULD
 validate that received data conforms to UTF-8.
 
+Stream Encoding Settings (``0x08``)
+-----------------------------------
+
+This frame type holds information defining the content encoding
+settings for a *stream*.
+
+This frame type is likely consumed by the protocol layer and is not
+passed on to applications.
+
+This frame type MUST ONLY occur on frames having the *Beginning of Stream*
+``Stream Flag`` set.
+
+The payload of this frame defines what content encoding has (possibly)
+been applied to the payloads of subsequent frames in this stream.
+
+The payload begins with an 8-bit integer defining the length of the
+encoding *profile*, followed by the string name of that profile, which
+must be an ASCII string. All bytes that follow can be used by that
+profile for supplemental settings definitions. See the section below
+on defined encoding profiles.
+
+Stream States and Flags
+-----------------------
+
+Streams can be in two states: *open* and *closed*. An *open* stream
+is active and frames attached to that stream could arrive at any time.
+A *closed* stream is not active. If a frame attached to a *closed*
+stream arrives, that frame MUST have an appropriate stream flag
+set indicating beginning of stream. All streams are in the *closed*
+state by default.
+
+The ``Stream Flags`` field denotes a set of bit flags for defining
+the relationship of this frame within a stream. The following flags
+are defined:
+
+0x01
+   Beginning of stream. The first frame in the stream MUST set this
+   flag. When received, the ``Stream ID`` this frame is attached to
+   becomes ``open``.
+
+0x02
+   End of stream. The last frame in a stream MUST set this flag. When
+   received, the ``Stream ID`` this frame is attached to becomes
+   ``closed``. Any content encoding context associated with this stream
+   can be destroyed after processing the payload of this frame.
+
+0x04
+   Apply content encoding. When set, any content encoding settings
+   defined by the stream should be applied when attempting to read
+   the frame. When not set, the frame payload isn't encoded.
+
+Streams
+-------
+
+Streams - along with ``Request IDs`` - facilitate grouping of frames.
+But the purpose of each is quite different and the groupings they
+constitute are independent.
+
+A ``Request ID`` is essentially a tag. It tells you which logical
+request a frame is associated with.
+
+A *stream* is a sequence of frames grouped for the express purpose
+of applying a stateful encoding or for denoting sub-groups of frames.
+
+Unlike ``Request ID``s which span the request and response, a stream
+is unidirectional and stream IDs are independent from client to
+server.
+
+There is no strict hierarchical relationship between ``Request IDs``
+and *streams*. A stream can contain frames having multiple
+``Request IDs``. Frames belonging to the same ``Request ID`` can
+span multiple streams.
+
+One goal of streams is to facilitate content encoding. A stream can
+define an encoding to be applied to frame payloads. For example, the
+payload transmitted over the wire may contain output from a
+zstandard compression operation and the receiving end may decompress
+that payload to obtain the original data.
+
+The other goal of streams is to facilitate concurrent execution. For
+example, a server could spawn 4 threads to service a request that can
+be easily parallelized. Each of those 4 threads could write into its
+own stream. Those streams could then in turn be delivered to 4 threads
+on the receiving end, with each thread consuming its stream in near
+isolation. The *main* thread on both ends merely does I/O and
+encodes/decodes frame headers: the bulk of the work is done by worker
+threads.
+
+In addition, since content encoding is defined per stream, each
+*worker thread* could perform potentially CPU bound work concurrently
+with other threads. This approach of applying encoding at the
+sub-protocol / stream level eliminates a potential resource constraint
+on the protocol stream as a whole (it is common for the throughput of
+a compression engine to be smaller than the throughput of a network).
+
+Having multiple streams - each with their own encoding settings - also
+facilitates the use of advanced data compression techniques. For
+example, a transmitter could see that it is generating data faster
+and slower than the receiving end is consuming it and adjust its
+compression settings to trade CPU for compression ratio accordingly.
+
+While streams can define a content encoding, not all frames within
+that stream must use that content encoding. This can be useful when
+data is being served from caches and being derived dynamically. A
+cache could pre-compressed data so the server doesn't have to
+recompress it. The ability to pick and choose which frames are
+compressed allows servers to easily send data to the wire without
+involving potentially expensive encoding overhead.
+
+Content Encoding Profiles
+-------------------------
+
+Streams can have named content encoding *profiles* associated with
+them. A profile defines a shared understanding of content encoding
+settings and behavior.
+
+The following profiles are defined:
+
+TBD
+
 Issuing Commands
 ----------------
 



To: indygreg, #hg-reviewers
Cc: mercurial-devel


More information about the Mercurial-devel mailing list
diff --git a/mercurial/wireprotoframing.py b/mercurial/wireprotoframing.py
--- a/mercurial/wireprotoframing.py
+++ b/mercurial/wireprotoframing.py
@@ -22,16 +22,27 @@
     util,
 )
 
-FRAME_HEADER_SIZE = 6
+FRAME_HEADER_SIZE = 8
 DEFAULT_MAX_FRAME_SIZE = 32768
 
+STREAM_FLAG_BEGIN_STREAM = 0x01
+STREAM_FLAG_END_STREAM = 0x02
+STREAM_FLAG_ENCODING_APPLIED = 0x04
+
+STREAM_FLAGS = {
+    b'stream-begin': STREAM_FLAG_BEGIN_STREAM,
+    b'stream-end': STREAM_FLAG_END_STREAM,
+    b'encoded': STREAM_FLAG_ENCODING_APPLIED,
+}
+
 FRAME_TYPE_COMMAND_NAME = 0x01
 FRAME_TYPE_COMMAND_ARGUMENT = 0x02
 FRAME_TYPE_COMMAND_DATA = 0x03
 FRAME_TYPE_BYTES_RESPONSE = 0x04
 FRAME_TYPE_ERROR_RESPONSE = 0x05
 FRAME_TYPE_TEXT_OUTPUT = 0x06
 FRAME_TYPE_PROGRESS = 0x07
+FRAME_TYPE_STREAM_SETTINGS = 0x08
 
 FRAME_TYPES = {
     b'command-name': FRAME_TYPE_COMMAND_NAME,
@@ -41,6 +52,7 @@
     b'error-response': FRAME_TYPE_ERROR_RESPONSE,
     b'text-output': FRAME_TYPE_TEXT_OUTPUT,
     b'progress': FRAME_TYPE_PROGRESS,
+    b'stream-settings': FRAME_TYPE_STREAM_SETTINGS,
 }
 
 FLAG_COMMAND_NAME_EOS = 0x01
@@ -94,6 +106,7 @@
     FRAME_TYPE_ERROR_RESPONSE: FLAGS_ERROR_RESPONSE,
     FRAME_TYPE_TEXT_OUTPUT: {},
     FRAME_TYPE_PROGRESS: {},
+    FRAME_TYPE_STREAM_SETTINGS: {},
 }
 
 ARGUMENT_FRAME_HEADER = struct.Struct(r'<HH')
@@ -104,55 +117,71 @@
 
     length = attr.ib()
     requestid = attr.ib()
+    streamid = attr.ib()
+    streamflags = attr.ib()
     typeid = attr.ib()
     flags = attr.ib()
 
 @attr.s(slots=True)
 class frame(object):
     """Represents a parsed frame."""
 
     requestid = attr.ib()
+    streamid = attr.ib()
+    streamflags = attr.ib()
     typeid = attr.ib()
     flags = attr.ib()
     payload = attr.ib()
 
-def makeframe(requestid, typeid, flags, payload):
+def makeframe(requestid, streamid, streamflags, typeid, flags, payload):
     """Assemble a frame into a byte array."""
     # TODO assert size of payload.
     frame = bytearray(FRAME_HEADER_SIZE + len(payload))
 
     # 24 bits length
     # 16 bits request id
+    # 8 bits stream id
+    # 8 bits stream flags
     # 4 bits type
     # 4 bits flags
 
     l = struct.pack(r'<I', len(payload))
     frame[0:3] = l[0:3]
-    struct.pack_into(r'<H', frame, 3, requestid)
-    frame[5] = (typeid << 4) | flags
-    frame[6:] = payload
+    struct.pack_into(r'<HBB', frame, 3, requestid, streamid, streamflags)
+    frame[7] = (typeid << 4) | flags
+    frame[8:] = payload
 
     return frame
 
 def makeframefromhumanstring(s):
     """Create a frame from a human readable string
 
     Strings have the form:
 
-        <request-id> <type> <flags> <payload>
+        <request-id> <stream-id> <stream-flags> <type> <flags> <payload>
 
     This can be used by user-facing applications and tests for creating
     frames easily without having to type out a bunch of constants.
 
-    Request ID is an integer.
+    Request ID and stream IDs are integers.
 
-    Frame type and flags can be specified by integer or named constant.
+    Stream flags, frame type, and flags can be specified by integer or
+    named constant.
 
     Flags can be delimited by `|` to bitwise OR them together.
     """
-    requestid, frametype, frameflags, payload = s.split(b' ', 3)
+    fields = s.split(b' ', 5)
+    requestid, streamid, streamflags, frametype, frameflags, payload = fields
 
     requestid = int(requestid)
+    streamid = int(streamid)
+
+    finalstreamflags = 0
+    for flag in streamflags.split(b'|'):
+        if flag in STREAM_FLAGS:
+            finalstreamflags |= STREAM_FLAGS[flag]
+        else:
+            finalstreamflags |= int(flag)
 
     if frametype in FRAME_TYPES:
         frametype = FRAME_TYPES[frametype]
@@ -169,7 +198,8 @@
 
     payload = util.unescapestr(payload)
 
-    return makeframe(requestid=requestid, typeid=frametype,
+    return makeframe(requestid=requestid, streamid=streamid,
+                     streamflags=finalstreamflags, typeid=frametype,
                      flags=finalflags, payload=payload)
 
 def parseheader(data):
@@ -179,17 +209,21 @@
     buffer is expected to be large enough to hold a full header.
     """
     # 24 bits payload length (little endian)
+    # 16 bits request ID
+    # 8 bits stream ID
+    # 8 bits stream flags
     # 4 bits frame type
     # 4 bits frame flags
     # ... payload
     framelength = data[0] + 256 * data[1] + 16384 * data[2]
-    requestid = struct.unpack_from(r'<H', data, 3)[0]
-    typeflags = data[5]
+    requestid, streamid, streamflags = struct.unpack_from(r'<HBB', data, 3)
+    typeflags = data[7]
 
     frametype = (typeflags & 0xf0) >> 4
     frameflags = typeflags & 0x0f
 
-    return frameheader(framelength, requestid, frametype, frameflags)
+    return frameheader(framelength, requestid, streamid, streamflags,
+                       frametype, frameflags)
 
 def readframe(fh):
     """Read a unified framing protocol frame from a file object.
@@ -216,7 +250,8 @@
         raise error.Abort(_('frame length error: expected %d; got %d') %
                           (h.length, len(payload)))
 
-    return frame(h.requestid, h.typeid, h.flags, payload)
+    return frame(h.requestid, h.streamid, h.streamflags, h.typeid, h.flags,
+                 payload)
 
 def createcommandframes(stream, requestid, cmd, args, datafh=None):
     """Create frames necessary to transmit a request to run a command.
@@ -398,12 +433,28 @@
 class stream(object):
     """Represents a logical unidirectional series of frames."""
 
+    def __init__(self, streamid, active=False):
+        self.streamid = streamid
+        self._active = False
+
     def makeframe(self, requestid, typeid, flags, payload):
         """Create a frame to be sent out over this stream.
 
         Only returns the frame instance. Does not actually send it.
         """
-        return makeframe(requestid, typeid, flags, payload)
+        streamflags = 0
+        if not self._active:
+            streamflags |= STREAM_FLAG_BEGIN_STREAM
+            self._active = True
+
+        return makeframe(requestid, self.streamid, streamflags, typeid, flags,
+                         payload)
+
+def ensureserverstream(stream):
+    if stream.streamid % 2:
+        raise error.ProgrammingError('server should only write to even '
+                                     'numbered streams; %d is not even' %
+                                     stream.streamid)
 
 class serverreactor(object):
     """Holds state of a server handling frame-based protocol requests.
@@ -483,6 +534,8 @@
         self._deferoutput = deferoutput
         self._state = 'idle'
         self._bufferedframegens = []
+        # stream id -> stream instance for all active streams from the client.
+        self._incomingstreams = {}
         # request id -> dict of commands that are actively being received.
         self._receivingcommands = {}
         # Request IDs that have been received and are actively being processed.
@@ -496,6 +549,30 @@
         Returns a dict with an ``action`` key that details what action,
         if any, the consumer should take next.
         """
+        if not frame.streamid % 2:
+            self._state = 'errored'
+            return self._makeerrorresult(
+                _('received frame with even numbered stream ID: %d' %
+                  frame.streamid))
+
+        if frame.streamid not in self._incomingstreams:
+            if not frame.streamflags & STREAM_FLAG_BEGIN_STREAM:
+                self._state = 'errored'
+                return self._makeerrorresult(
+                    _('received frame on unknown inactive stream without '
+                      'beginning of stream flag set'))
+
+            self._incomingstreams[frame.streamid] = stream(frame.streamid)
+
+        if frame.streamflags & STREAM_FLAG_ENCODING_APPLIED:
+            # TODO handle decoding frames
+            self._state = 'errored'
+            raise error.ProgrammingError('support for decoding stream payloads '
+                                         'not yet implemented')
+
+        if frame.streamflags & STREAM_FLAG_END_STREAM:
+            del self._incomingstreams[frame.streamid]
+
         handlers = {
             'idle': self._onframeidle,
             'command-receiving': self._onframecommandreceiving,
@@ -513,6 +590,8 @@
 
         The raw bytes response is passed as an argument.
         """
+        ensureserverstream(stream)
+
         def sendframes():
             for frame in createbytesresponseframesfrombytes(stream, requestid,
                                                             data):
@@ -552,6 +631,8 @@
         }
 
     def onapplicationerror(self, stream, requestid, msg):
+        ensureserverstream(stream)
+
         return 'sendframes', {
             'framegen': createerrorframe(stream, requestid, msg,
                                          application=True),